"""
Tracks the message data and exports it as a renderable table.

@date: 03.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from typing import Final, Iterable, ClassVar, Any, NamedTuple
from dataclasses import dataclass
from math import ceil

# 3rd-party
from rich.table import Table
from rich.pretty import Pretty
from rich.box import DOUBLE
from exhausterr import Result, Ok, Err, Error


# Local
from ._monitor import UnknownMessage, DecodedMessage, CanMessage


class CsvRecord(NamedTuple):
    """
    Structure for CSV records in plots
    """

    timestamp: float
    value: Any


@dataclass
class InvalidName(Error):
    exception_cls: ClassVar[type[Exception]] = NameError
    name: str


@dataclass
class InvalidType(Error):
    exception_cls: ClassVar[type[Exception]] = ValueError
    value: Any


DEFAULT_WIDTH: Final[int] = 200
DEFAULT_HEIGHT: Final[int] = 40


class MessageTable:
    """
    Stateful object; tracks the latest version of each message received
    and provides a primitive to export the current state as single renderable table.
    Messages are sorted by IDs.
    """

    def __init__(
        self,
        ignore_unknown_messages: bool = False,
        filters: Iterable[int | str] = (),
    ) -> None:
        """
        Parameters
        ----------
        ignore_unknown_messages: bool
            If enabled, does not include unknown messages in the exported table.
        """
        self._ignore_unknown_messages = ignore_unknown_messages
        # CAN ID is used as primary key
        self._decoded_messages: dict[int, DecodedMessage] = {}
        self._raw_messages: dict[int, CanMessage] = {}
        self._id_filters = set((f for f in filters if isinstance(f, int)))
        self._name_filters = set((f for f in filters if isinstance(f, str)))
        self._plots: dict[str, dict[str, list[CsvRecord]]] = {}
        self._page_height: int = DEFAULT_WIDTH
        self._page_width: int = DEFAULT_WIDTH

    def set_page_dimensions(self, width: int | None, height: int | None) -> int:
        """
        Returns
        -------
        int
            Number of pages required to render the entire data
        """
        self._page_height = height or self._page_height
        self._page_width = width or self._page_width
        return ceil(self.renderable_size() / self._page_height)

    def renderable_size(self) -> int:
        """
        Returns
        -------
        int
            Number of lines required to render the entire table
        """
        length = 0
        for decoded in self._decoded_messages.values():
            if self.filter_message_id(decoded.can_id) or self.filter_message_name(
                decoded.message_name
            ):
                continue
            length += len(decoded.data) + 2  # Need extra lines for brackets

        if not self._ignore_unknown_messages:
            length += len(self._raw_messages)
        return length

    def start_plot(self, message_signal_key: str) -> Result[None, InvalidName]:
        """
        Starts recording values for the message signal pair given.
        Parameters
        ----------
        message_signal_key: str
            Given as `message_name.signal_name`

        Returns
        -------
        Result[None, InvalidName]
            Nothing if OK,
            InvalidName if the message signal pair is invalid
        """
        try:
            message, signal = message_signal_key.split(".")
        except ValueError:
            return Err(InvalidName(message_signal_key))
        self._plots.setdefault(message, {})[signal] = []
        return Ok(None)

    def take_snapshot(self) -> dict[str, Any]:
        """
        Exports the current data as a CSV snapshot
        """
        snapshot = {}
        for msg_name, decoded in self._decoded_messages.items():
            for signal_name, value in decoded.data.items():
                snapshot[f"{msg_name}.{signal_name}"] = value

        return snapshot

    def get_plot_by_name(
        self, message_signal_key: str
    ) -> Result[tuple[list[int], list[float]], InvalidName]:
        """
        Returns
        -------
        Result[tuple[list[int], list[float]], InvalidName]
            If OK, the tuple of x and y values.
            Else InvalidName if message-signal key is invalid
        """
        try:
            message, signal = message_signal_key.split(".")
        except ValueError:
            return Err(InvalidName(message_signal_key))

        try:
            values = self._plots[message][signal]
        except KeyError as exc:
            Err(InvalidName(str(exc)))

        x_vals = []
        y_vals = []
        for x, y in values:
            x_vals.append(x)
            y_vals.append(y)
        return Ok((x_vals, y_vals))

    def export_plots_to_csv(self) -> list[str]:
        """
        Exports all plots to CSV files

        Returns
        -------
        str
            Path to the created CSV file
        """
        csv_paths = []
        for message, signals in self._plots.items():
            for signal, values in signals.items():
                csv_name = f"{message}.{signal}.csv"
                csv_paths.append(csv_name)
                with open(csv_name, "w") as f:
                    f.write(f"timestamp,{message}.{signal}\n")
                    for record in values:
                        f.write(",".join([str(v) for v in record]))
                        f.write("\n")
        return csv_paths

    def filter_message_id(self, can_id: int) -> bool:
        """
        Returns
        -------
        bool
            Whether the message should be filtered out
        """
        if not self._id_filters:
            return False

        return can_id not in self._id_filters

    def filter_message_name(self, name: str) -> bool:
        """
        Returns
        -------
        bool
            Whether the message should be filtered out
        """
        if not self._name_filters:
            return False

        return name not in self._name_filters

    def update(self, message: Result[DecodedMessage, UnknownMessage]) -> None:
        """
        Updates the internal table based on the given result.
        """
        # overriding the last version or creating a new one if first encounter
        match message:
            case Ok(decoded):
                self._decoded_messages[decoded.can_id] = decoded
                self._update_plots(decoded)

            case Err(UnknownMessage(can_id, raw_msg)):
                self._raw_messages[can_id] = raw_msg

    def _update_plots(self, message: DecodedMessage) -> Result[None, InvalidType]:
        """
        Updates all plots that are currently current for the passed
        decoded message.
        """
        message_plot_dict = self._plots.get(message.message_name)
        if message_plot_dict is None:
            return Ok(None)

        for signal, buffer in message_plot_dict.items():
            received = message.data[signal]
            try:
                new_value = float(message.data[signal])
            except ValueError:
                return Err(InvalidType(received))
            buffer.append(CsvRecord(message.timestamp, new_value))
        return Ok(None)

    def _format_binary_data(self, data: bytes | bytearray) -> str:
        """
        A simple string formatter for binary data
        """
        return " ".join((f"{b:02X}" for b in data))

    def _table_builder(self) -> Table:
        """
        Default builder for rich's Table
        """
        return Table(
            title="Messages",
            min_width=self._page_width,
            width=self._page_width,
            expand=True,
            box=DOUBLE,
            header_style="bold cyan",
            title_style="bold underline green",
        )

    def export_single_message(self, message_id: int | str) -> Table | None:
        decoded: DecodedMessage | None = None
        if isinstance(message_id, str):
            for decoded_msg in self._decoded_messages.values():
                if decoded_msg.message_name == message_id:
                    decoded = decoded_msg
                    break
            else:
                return None
        else:
            decoded = self._decoded_messages.get(message_id)

        if decoded is None:
            return None
        table = self._table_builder()
        table.add_column("ID", justify="right", style="cyan")
        table.add_column("Name", style="yellow")
        table.add_column("Binary", style="green")
        table.add_column("Decoded", style="blue")
        for signal in decoded.data:
            table.add_column(signal, style="rgb(128,128,128)")

        table.add_row(
            f"{decoded.can_id:08X}",
            str(decoded.message_name),
            self._format_binary_data(decoded.binary),
            *(Pretty(val) for val in decoded.data.values()),
        )

        return table

    def export_paginated(self, page_index: int = 0) -> Table:
        """
        Returns
        -------
        Table
            The current tracked data as a renderable table.
        """
        page_starts = self._page_height * page_index
        page_ends = page_starts + self._page_height
        current_index = 0
        table = self._table_builder()
        table.add_column("ID", justify="right", style="cyan")
        table.add_column("Name", style="yellow")
        table.add_column("Binary", style="green")
        table.add_column("Decoded", style="blue")

        for decoded in self._decoded_messages.values():
            if current_index > page_ends:
                return table
            if any(
                (
                    self.filter_message_id(decoded.can_id),
                    self.filter_message_name(decoded.message_name),
                )
            ):
                continue
            current_index += len(decoded.data) + 2
            if current_index < page_starts:
                continue

            # Pre-formatting floating numbers to prevent GUI wiggling
            # due to tiny floating-point precision inconsistencies thingies
            preformatted_data = {
                key: f"{val:.2f}" if isinstance(val, float) else str(val)
                for key, val in decoded.data.items()
            }
            table.add_row(
                f"{decoded.can_id:08X}",
                str(decoded.message_name),
                self._format_binary_data(decoded.binary),
                Pretty(preformatted_data),
            )

        if self._ignore_unknown_messages:
            return table
        for raw_msg in self._raw_messages.values():
            if current_index > page_ends:
                return table
            if self.filter_message_id(raw_msg.arbitration_id):
                continue

            current_index += 1
            if current_index < page_starts:
                continue
            table.add_row(
                f"{raw_msg.arbitration_id:08X}",
                "[Unknown]",
                self._format_binary_data(raw_msg.data),
                "",
            )
        return table



"""
Application entrypoint(s) for this package.
CLI tools are based on click and will be installed
automatically as package scripts.

@date: 04.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations

import asyncio
import json
import os
import sys
import time
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, auto

# built-in
from typing import IO, Callable, Iterable, Iterator, Literal

import can
import cantools

# 3rd-party
import click
from cantools.database.can import Database
from exhausterr import Err, Ok
from rich.console import Console, Group
from rich.live import Live
import asciichartpy as acp

from ._console import MessageTable

# Local
from ._monitor import (
    CanMonitor,
    get_platform_default_channel,
    get_platform_default_driver,
)
from ._utils import CanIdPattern, InvalidPattern, convert_pattern_to_mask

# Number of lines for the actual table
HEIGHT_MARGIN: int = 10
WIDTH_MARGIN: int = 10
DEFAULT_HEIGHT = 30
ZOOM_FACTOR = 1.1
PLOT_MAX_SIZE = 50


class UserCommands(Enum):
    """
    All the commands which can be received from the user
    """

    TAKE_SNAPSHOT = auto()


@dataclass
class UserInterface:
    """
    Manages the interaction with the user
    """

    page_index: int = 0
    total_pages: int = 1
    dispatcher: dict[UserCommands, Callable[[], None]] = field(default_factory=dict)
    log: str = ""
    height: int = DEFAULT_HEIGHT

    def on_input(self, stream: IO[str]) -> None:
        """
        Process user input from the given stream
        """
        # clearing previous log
        self.log = ""
        command = stream.readline().strip()
        match command:
            case "":
                self.page_index = (self.page_index + 1) % self.total_pages

            case "b":
                self.page_index = (self.page_index - 1) % self.total_pages

            case "s":
                if (cmd := self.dispatcher.get(UserCommands.TAKE_SNAPSHOT)) is not None:
                    match cmd():
                        case Ok(_):
                            self.log = "[green]Took snapshot successfully !"
                        case Err(err):
                            self.log = str(err)

            # zoom in / zoom out commands
            case "+" | "++" | "+++":
                zoom_factor = ZOOM_FACTOR ** len(command)
                self.height = round(self.height / zoom_factor)

            case "-" | "--" | "---":
                zoom_factor = ZOOM_FACTOR ** len(command)
                self.height = round(self.height * zoom_factor)

            case _:
                if command.isnumeric():
                    try:
                        idx = int(command) - 1
                    except ValueError:
                        # ignoring
                        return

                    if 0 <= idx < self.total_pages:
                        self.page_index = idx
                else:
                    self.log = f"[red]Unknown command: {command}"

    def page_indication(self) -> str:
        return (
            f"Page {self.page_index + 1}/{self.total_pages}"
            " (Press enter go to next page)"
        )


async def _canviewer(
    channel: str,
    driver: str,
    databases: Iterable[Database],
    ignore_unknown_messages: bool,
    message_filters: Iterable[int | str],
    single_message: str | None = None,
    record_signals: list[str] = [],
    plot_signals: list[str] = [],
    inline: bool = False,
    snapshot_type: Literal["csv", "json"] = "csv",
    mask: int = 0xFFFF_FFFF,
    id_pattern: CanIdPattern | int | None = None,
) -> None:
    """
    Main asynchronous runner for the console application.

    Parameters
    ----------
    channel: str
        The name of the CAN channel to monitor

    driver: str
        The name of the CAN driver (interface)

    databases: Iterable[Database]
        The paths to .kcd files or to a folder containing kcd files
    """
    message_table = MessageTable(
        ignore_unknown_messages=ignore_unknown_messages, filters=message_filters
    )
    console = Console()
    loop = asyncio.get_event_loop()
    started = datetime.now()

    def on_snapshot() -> Ok[None]:
        dict_data = message_table.take_snapshot()
        fname = (
            "snapshot_canviewer_"
            + started.strftime("%Y_%m_%d_%H_%M_%S")
            + "."
            + snapshot_type
        )
        now = time.time()
        if not os.path.exists(fname) and snapshot_type == "csv":
            with open(fname, "w+") as f:
                header = ",".join(("timestamp", *dict_data.keys()))
                f.write(header + "\n")
        # converting to CSV
        with open(fname, "a+") as f:
            match snapshot_type:
                case "json":
                    dict_data["timestamp"] = now
                    f.write(json.dumps(dict_data, default=str, indent=4))

                case "csv":
                    values = ",".join((str(now), *[str(v) for v in dict_data.values()]))
                    f.write(values + "\n")
        return Ok()

    for message_signal in (*record_signals, *plot_signals):
        if not message_table.start_plot(message_signal):
            click.echo(f"Invalid message signal: {message_signal}")
            return

    with can.Bus(interface=driver, channel=channel) as bus:
        with Live(console=console, screen=not inline) as live:
            interface = UserInterface(height=console.size.height)

            # registering commands
            interface.dispatcher[UserCommands.TAKE_SNAPSHOT] = on_snapshot
            loop.add_reader(sys.stdin, interface.on_input, sys.stdin)
            backend = CanMonitor(bus, *databases, mask=mask, id_pattern=id_pattern)
            try:
                while True:  # Ctrl + C to leave
                    message = await backend.queue.get()
                    message_table.update(message)

                    page_height = interface.height - HEIGHT_MARGIN
                    page_width = console.size.width - WIDTH_MARGIN
                    interface.total_pages = message_table.set_page_dimensions(
                        page_width, page_height
                    )
                    if single_message is not None:
                        renderable_table = message_table.export_single_message(
                            single_message
                        )
                        if renderable_table is None:
                            continue
                    else:
                        renderable_table = message_table.export_paginated(
                            interface.page_index
                        )

                    plots = []
                    for message_signal in plot_signals:
                        match message_table.get_plot_by_name(message_signal):
                            case Ok(data):
                                _, y = data
                                plots.append(
                                    Group(
                                        message_signal,
                                        acp.plot(y[:PLOT_MAX_SIZE]),
                                    )
                                )

                            case Err(error):
                                interface.log = f"[red]{error}"

                    if interface.total_pages > 1 and single_message is None:
                        renderable = Group(
                            interface.page_indication(),
                            interface.log,
                            *plots,
                            renderable_table,
                        )
                    else:
                        renderable = Group(*plots, renderable_table)

                    live.update(renderable)

            finally:
                csv_paths = message_table.export_plots_to_csv()
                if csv_paths:
                    click.echo(f"CSV files created: {csv_paths}")


def collect_databases(*paths: str) -> Iterator[str]:
    """
    Based on a selection or folder and/or direct paths to
    CAN databases, discovers the databases within the folders
    and yields a flattened list of all discovered databases.

    Yields
    ------
    str
        Path to a KCD or DBC database to loadd
    """
    for path in paths:
        if os.path.isdir(path):
            yield from (
                os.path.join(path, f)
                for f in os.listdir(path)
                if f.endswith(".kcd") or f.endswith(".dbc")
            )
        else:
            yield path


@click.command()
@click.option(
    "-c", "--channel", default=None, type=str, help="Name of the CAN channel to monitor"
)
@click.option(
    "-d",
    "--driver",
    default=None,
    type=str,
    help="Specifies which CAN driver to use if multiple available",
)
@click.option(
    "-db",
    "--databases",
    default=(),
    type=str,
    multiple=True,
    help="Path to .kcd files or to a folder containing kcd files",
)
@click.option(
    "-f",
    "--filters",
    default=(),
    type=str,
    multiple=True,
    help="Either a name or a numeric ID, only passed messages will be displayed",
)
@click.option(
    "-s",
    "--single-message",
    type=str,
    default=None,
    help="Tracks a single message, shows a custom table with one column per signal",
)
@click.option(
    "-i",
    "--ignore-unknown-messages",
    is_flag=True,
    help="Hides messages that are not declared in one of your databases",
)
@click.option(
    "-r",
    "--record-signals",
    multiple=True,
    type=str,
    help=(
        "Records the values for a given signal, exports them to CSV on exiting.\n"
        "If you'd like to also plot the signal in realtime, use -pl/--plot instead\n"
        "You shall pass your target signal as message_name.signal_name"
    ),
)
@click.option(
    "-pl",
    "--plot-signals",
    multiple=True,
    type=str,
    help=(
        "Plots the values for a given signal, exports them to CSV on exiting.\n"
        "You shall pass your target signal as message_name.signal_name"
    ),
)
@click.option(
    "-n",
    "--inline",
    is_flag=True,
    help="Disables full-screen",
)
@click.option(
    "-sf",
    "--snapshot-format",
    type=click.Choice(["json", "csv"]),
    default="csv",
    help="Format to use for snapshots",
)
@click.option(
    "-mk",
    "--mask",
    type=str,
    default="FFFFFFFF",
    help="Applies this mask on CAN IDs before feeding to decoder",
)
@click.option(
    "-p",
    "--pattern",
    type=str,
    default=None,
    help="Filter in all the messages following that pattern",
)
def canviewer(
    channel: str | None,
    driver: str | None,
    databases: Iterable[str],
    filters: Iterable[str],
    single_message: str | None,
    ignore_unknown_messages: bool,
    record_signals: list[str],
    plot_signals: list[str],
    inline: bool,
    snapshot_format: Literal["json", "csv"],
    mask: str,
    pattern: str | None,
) -> None:
    """
    For every CAN ID found on the CAN bus,
    displays the data for the last message received.
    If the message is declared in one of the passed databases,
    shows the decoded data.
    """

    try:
        id_pattern = convert_pattern_to_mask(pattern) if pattern else None
    except InvalidPattern as exc:
        raise click.BadParameter(str(exc), param_hint="pattern") from exc

    if channel is None:
        match get_platform_default_channel():
            case Ok(channel_name):
                channel = channel_name
            case Err(error):
                click.echo(str(error))
                return

    if driver is None:
        match get_platform_default_driver():
            case Ok(driver_name):
                driver = driver_name
            case Err(error):
                click.echo(str(error))
                return
    converted_filters: list[int | str] = [
        int(f) if f.isnumeric() else f for f in filters
    ]
    loaded_dbs: Iterable[Database] = map(
        cantools.database.load_file,  # type: ignore[arg-type]
        collect_databases(*databases),
    )
    asyncio.run(
        _canviewer(
            channel,
            driver,
            loaded_dbs,
            ignore_unknown_messages,
            converted_filters,
            single_message=single_message,
            record_signals=record_signals,
            plot_signals=plot_signals,
            inline=inline,
            snapshot_type=snapshot_format,
            id_pattern=id_pattern,
            mask=int(mask, 16),
        )
    )



"""
Test suite for shared utilities

@date: 05.11.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
import pytest
from canviewer import CanIdPattern, convert_pattern_to_mask
from canviewer._utils import InvalidPattern


@pytest.mark.parametrize(
    "pattern, value, expected",
    [
        [(0x0120_0000, 0xFFF0_0000), 0x0123_4567, True],
        [(0x1120_0000, 0xFFF0_0000), 0x01234567, False],
        [(0x0023_4000, 0x00FF_F000), 0x01234567, True],
        [(0x0023_0000, 0x00FF_F000), 0x01234567, False],
    ],
)
def test_pattern_matching(pattern: tuple[int, int], value: int, expected: bool) -> None:
    """
    Verifies the `match` method of `CanIdPattern`.
    """
    id_pattern = CanIdPattern(*pattern)
    assert id_pattern.match(value) == expected


@pytest.mark.parametrize(
    "pattern,expected",
    [
        ("1234", 0x1234),
        ("*1234", CanIdPattern(0x1234, 0x0000_FFFF)),
        ("1234*", CanIdPattern(0x1234_0000, 0xFFFF_0000)),
        ("12345*", CanIdPattern(0x1234_5000, 0xFFFF_F000)),
        ("12*", CanIdPattern(0x1200_0000, 0xFF00_0000)),
        ("1234$", InvalidPattern),
        ("12$4*", InvalidPattern),
        ("*12$4", InvalidPattern),
    ],
)
def test_convert_pattern_to_mask(
    pattern: str, expected: CanIdPattern | int | type[Exception]
) -> None:
    """
    Verifies the `convert_pattern_to_mask` function.
    """
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            convert_pattern_to_mask(pattern)
    else:
        obtained = convert_pattern_to_mask(pattern)
        assert obtained == expected, obtained



"""
Miscaellenous utilities

@date: 05.11.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from dataclasses import dataclass


class InvalidPattern(Exception):
    """
    Whenever the pattern passed by the user is invalid
    """

    pass


@dataclass(eq=True)
class CanIdPattern:
    """
    A match pattern for a CAN message.

    Bits captured by the `mask` are compared to the ones
    in `values`.
    Can ID latches the pattern if (can_id & mask) == value
    """

    value: int
    mask: int = 0xFFFF_FFFF

    def match(self, can_id: int) -> bool:
        """
        Returns
        -------
        bool
            True if the CAN ID should be filtered in,
            False for filtering out
        """
        return (can_id & self.mask) == self.value

    def __str__(self) -> str:
        """
        Returns
        -------
        str
            Friendly representation that uses hex as default for integers
        """
        return f"0x{self.value:08x}; 0x{self.mask:08x}"

    def __repr__(self) -> str:
        """
        Returns
        -------
        str
            A human readable representation of the pattern
        """
        return f"Pattern: 0x{self.value:08x}; Mask: {self.mask:08x}"


def _convert_from_hex(pattern: str) -> int:
    """
    A small helper that raises the proper user exception when getting invalid
    pattern values
    """
    try:
        return int(pattern, 16)
    except ValueError as exc:
        raise InvalidPattern(
            f"Got invalid characters after `*` in pattern, expected hex digits: {pattern}"
        ) from exc


def convert_pattern_to_mask(pattern: str) -> CanIdPattern | int:
    """
    Converts a match pattern given by the user
    to a acceptance mask.

    Examples
    --------
    *123 should expand to 0x00000FFF, 0x00000123
    123* should expand to 0x00000FFF,0x12300000
    0000FFFF, 1234 should expand
    """
    if "," in pattern:
        try:
            mask, pattern = pattern.split(",")
        except ValueError as exc:
            raise InvalidPattern(
                f"Passing pattern with `,` expects only two values, got {pattern}"
            ) from exc
        return CanIdPattern(int(mask, 16), int(pattern, 16))

    if pattern.startswith("*"):
        pattern = pattern[1:]
        pattern_value = _convert_from_hex(pattern)
        # We pad based on 32 bits masks, hence 8 hex digits
        value_shift = len(pattern) * 4  # each digit is 4 bits
        mask = (1 << value_shift) - 1
        return CanIdPattern(pattern_value, mask)

    if pattern.endswith("*"):
        pattern = pattern[:-1]
        pattern_value = _convert_from_hex(pattern)
        # We pad based on 32 bits masks, hence 8 hex digits
        bitsize = len(pattern) * 4  # each digit is 4 bits
        value_shift = 32 - bitsize
        pattern_value <<= value_shift
        mask = (2**bitsize - 1) << value_shift
        return CanIdPattern(pattern_value, mask)

    return _convert_from_hex(pattern)



"""
Manages the interactions with the CAN bus.
Monitors the bus and atttempts to decode the received messages when possible.
Queues the result to make them available to the frontend.

@date: 04.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
import asyncio
from asyncio import Queue
from typing import Iterator, Union, ClassVar, cast
import platform
from dataclasses import dataclass
from exhausterr import Result, Ok, Err, Error
from can import Message as CanMessage, BusABC
from cantools.database.can import Database as CanDatabase  # type: ignore[attr-defined]
from cantools.database.can.message import Message as CanFrame
from cantools.database.namedsignalvalue import NamedSignalValue

from ._utils import CanIdPattern

# type hinting
CanTypes = Union[int, float, str, NamedSignalValue]
MessageDict = dict[str, CanTypes]


@dataclass
class UnknownMessage(Error):
    """
    Emitted when a message is not registered in any of the tracked databases
    """

    description: ClassVar[str | None] = (
        "CAN ID {can_id} is not registered in any of the tracked databases"
    )
    exception_cls: ClassVar[type[Exception]] = ValueError

    # error parameters
    can_id: int
    message: CanMessage

    def __hash__(self) -> int:
        """
        Uses the CAN Id as the main key
        """
        return self.can_id


@dataclass
class UnsupportedSystem(Error):
    """
    CAN operations are not supported on the target system
    """

    description: ClassVar[str | None] = "Unsupported system: {system}"
    exception_cls: ClassVar[type[Exception]] = NotImplementedError

    # error parameters
    system: str


@dataclass
class MuxSelectorValue:
    """
    A mux selector value
    """

    name: str
    value: CanTypes


@dataclass
class DecodedMessage:
    """
    Simple container for a decoded message,
    keeps track of the original CAN ID and message name
    as stated in the CAN database
    """

    can_id: int
    timestamp: float
    frame_name: str
    binary: bytearray
    data: MessageDict
    mux_selectors: tuple[MuxSelectorValue, ...] = ()

    def __hash__(self) -> int:
        """
        Uses the message as the main key
        """
        return hash(self.message_name)

    @property
    def message_name(self) -> str:
        """
        Returns
        -------
        str
            The name of the message, combining the frame name and the mux values.
        """
        formatted_selectors = (
            f"[{mux.name}={mux.value}]" for mux in self.mux_selectors
        )
        return self.frame_name + "".join(formatted_selectors)


def get_platform_default_channel() -> Result[str, UnsupportedSystem]:
    """
    Returns
    -------
    Result[str, UnsupportedSystem]
        The default channel to use for this system.
        UnsupportedSystem is the local platform is not supported
    """
    local_system = platform.system()
    match local_system:
        case "Windows":
            return Ok("PCAN_USBBUS1")
        case "Linux":
            return Ok("can0")
        case _:
            return Err(UnsupportedSystem(local_system))


def get_platform_default_driver() -> Result[str, UnsupportedSystem]:
    """
    Returns
    -------
    Result[str, UnsupportedSystem]
        The default driver to use for this system.
        UnsupportedSystem is the local platform is not supported
    """
    local_system = platform.system()
    match local_system:
        case "Windows":
            return Ok("pcan")
        case "Linux":
            return Ok("socketcan")
        case _:
            return Err(UnsupportedSystem(local_system))


class CanMonitor:
    """
    Monitors a bus and decodes received messages when possible
    """

    def __init__(
        self,
        bus: BusABC,
        *can_dbs: CanDatabase,
        loop: asyncio.AbstractEventLoop | None = None,
        mask: int = 0xFFFF_FFFF,
        id_pattern: CanIdPattern | int | None = None,
    ) -> None:
        self._loop = loop or asyncio.get_event_loop()
        self._queue: Queue[Result[DecodedMessage, UnknownMessage]] = Queue()
        self._dbs = list(can_dbs)
        self._bus = bus
        # Starting the monitor
        self._loop.add_reader(self._bus, self.handler)
        self._mask = mask
        if id_pattern is not None:
            self._id_pattern = (
                id_pattern
                if isinstance(id_pattern, CanIdPattern)
                else CanIdPattern(id_pattern, ~mask)
            )
        else:
            self._id_pattern = None

    @property
    def queue(self) -> Queue:
        """
        Returns
        -------
        Queue
            All the decoding results for received messages.
        """
        return self._queue

    def get_mux_selector_values(
        self, frame: CanFrame, data: MessageDict
    ) -> Iterator[MuxSelectorValue]:
        """
        Returns
        -------
        int | None
            The index of the mux if the frame is a mux, None otherwise.
        """
        if frame.signal_tree is None:
            return
        for entry in frame.signal_tree:
            if not isinstance(entry, dict):
                # non-mux are just normal names, not dict
                continue
            for mux_name in entry:
                selected_value = data.get(mux_name)
                if selected_value is None:
                    continue

                yield MuxSelectorValue(mux_name, selected_value)

    def decode_message(self, msg: CanMessage) -> Result[DecodedMessage, UnknownMessage]:
        """
        Looks for a matching message in the list of tracked databases.
        Decodes the message if possible otherwise returns a `UnknownMessage`
        with the received message data
        """
        can_id = msg.arbitration_id
        # applying mask
        candidate_id = can_id & self._mask
        for db in self._dbs:
            try:
                frame = db.get_message_by_frame_id(candidate_id)
                decoded_data = frame.decode(msg.data)  # type: ignore[assignment]
                # Have to cast because cantools does not provide necessary overloads
                # for decode -> when decode_containers is False, returned type is dict
                decoded_data = cast(MessageDict, decoded_data)
                # checking if frame is a mux

                selectors = tuple(self.get_mux_selector_values(frame, decoded_data))
                decoded_msg = DecodedMessage(
                    can_id=can_id,
                    timestamp=msg.timestamp,
                    frame_name=frame.name,
                    binary=msg.data,
                    data=cast(MessageDict, decoded_data),
                    mux_selectors=selectors,
                )

                return Ok(decoded_msg)
            except KeyError:
                continue

        return Err(UnknownMessage(candidate_id, msg))

    def handler(self) -> None:
        """
        Main callback on message reception.
        Attempts decoding the received data and queues the result.
        """
        next_message = self._bus.recv(timeout=0.0)

        if next_message is None:
            return

        can_id = next_message.arbitration_id
        if self._id_pattern is not None:
            if not (self._id_pattern.match(can_id)):
                return

        self._queue.put_nowait(self.decode_message(next_message))
